# django-mojeid-auth -  MojeID integration for django.contrib.auth
#
# Copyright (C) 2013-2015 CZ.NIC
# Copyright (C) 2008-2013 Canonical Ltd.
# Copyright (C) 2007 Simon Willison
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import time

try:
    # python3
    from urllib.parse import urlsplit
except ImportError:
    # python2
    from urlparse import urlsplit

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.urlresolvers import reverse
from django.http import Http404, HttpResponse, HttpResponseRedirect
from django.shortcuts import render_to_response
from django.template import RequestContext
from django.template.loader import render_to_string
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST

from openid.consumer.consumer import SUCCESS, CANCEL, FAILURE
from openid.cryptutil import randomString
from openid.extensions import ax, pape
from openid.fetchers import HTTPFetchingError
from openid.kvform import dictToKV
from openid.message import Message
from openid.store.nonce import NONCE_CHARS
from openid.yadis.constants import YADIS_CONTENT_TYPE

from django_mojeid import errors
from django_mojeid.auth import OpenIDBackend

from django_mojeid.exceptions import (
    DjangoOpenIDException,
    IdentityAlreadyClaimed,
)
from django_mojeid.models import Nonce, UserOpenID
from django_mojeid.mojeid import (
    Assertion,
    get_attributes,
    get_attribute_query,
    get_registration_url,
    create_service,
    MojeIDConsumer
)
from django_mojeid.settings import mojeid_settings
from django_mojeid.signals import (
    user_login_report,
    trigger_error,
    authenticate_user,
    associate_user
)
from django_mojeid.store import DjangoOpenIDStore


SESSION_ATTR_SET_KEY = 'mojeid_attr_set'


def sanitise_redirect_url(redirect_to):
    """Sanitise the redirection URL."""
    # Light security check -- make sure redirect_to isn't garbage.
    is_valid = True
    if not redirect_to or ' ' in redirect_to:
        is_valid = False
    elif '//' in redirect_to:
        # Allow the redirect URL to be external if it's a permitted domain
        allowed_domains = getattr(settings, "ALLOWED_EXTERNAL_OPENID_REDIRECT_DOMAINS", [])
        s, netloc, p, q, f = urlsplit(redirect_to)
        # allow it if netloc is blank or if the domain is allowed
        if netloc:
            # a domain was specified. Is it an allowed domain?
            if netloc.find(":") != -1:
                netloc, _ = netloc.split(":", 1)
            if netloc not in allowed_domains:
                is_valid = False

    # If the return_to URL is not valid, use the default.
    if not is_valid:
        redirect_to = settings.LOGIN_REDIRECT_URL

    return redirect_to


def render_failure(request, error, template_name='openid/failure.html'):
    """Render an error page to the user."""
    # Render the response to trigger_error signal
    resp = trigger_error.send(sender=__name__, error=error, request=request)
    # Return first valid response
    for r in resp:
        if r[1] is not None and isinstance(r[1], HttpResponse):
            return r[1]

    # No response to signal - render default page
    data = render_to_string(template_name, {'message': error.msg},
                            context_instance=RequestContext(request))
    return HttpResponse(data, status=error.http_status)


@require_POST
def login_begin(request, attribute_set='default'):
    """Begin an MojeID login request."""

    if mojeid_settings.MOJEID_SESSION_NEXT_PAGE_ATTR in request.session:
        del request.session[mojeid_settings.MOJEID_SESSION_NEXT_PAGE_ATTR]

    # create consumer, start login process
    consumer = MojeIDConsumer(DjangoOpenIDStore())
    openid_request = consumer.begin(create_service())

    # Request user details.
    attributes = get_attribute_query(attribute_set)
    # save settings set name for response handler
    request.session[SESSION_ATTR_SET_KEY] = attribute_set

    fetch_request = ax.FetchRequest()
    for attribute, required in attributes:
        fetch_request.add(attribute.generate_ax_attrinfo(required))

    if attributes:
        openid_request.addExtension(fetch_request)

    if mojeid_settings.MOJEID_LOGIN_METHOD != 'ANY' or \
            mojeid_settings.MOJEID_MAX_AUTH_AGE is not None:
        # set authentication method to OTP or CERT
        if mojeid_settings.MOJEID_LOGIN_METHOD == "OTP":
            auth_method = [pape.AUTH_MULTI_FACTOR]
        elif mojeid_settings.MOJEID_LOGIN_METHOD == "CERT":
            auth_method = [pape.AUTH_PHISHING_RESISTANT]
        else:
            auth_method = None

        pape_request = pape.Request(
            preferred_auth_policies=auth_method,
            max_auth_age=mojeid_settings.MOJEID_MAX_AUTH_AGE,
        )
        openid_request.addExtension(pape_request)

    # Construct the request completion URL
    return_to = request.build_absolute_uri(reverse(login_complete))

    # get 'next page' and save it to the session
    redirect_to = sanitise_redirect_url(OpenIDBackend.get_redirect_to(request))
    if redirect_to:
        request.session[mojeid_settings.MOJEID_SESSION_NEXT_PAGE_ATTR] = redirect_to

    # Realm should be always something like 'https://example.org/openid/'
    realm = getattr(settings, 'MOJEID_REALM', None)
    if not realm:
        realm = request.build_absolute_uri(reverse(top))

    # we always use POST request
    form_html = openid_request.htmlMarkup(
        realm, return_to, form_tag_attrs={'id': 'openid_message'})
    return HttpResponse(form_html, content_type='text/html; charset=UTF-8')


def registration(request, attribute_set='default',
                 template_name='openid/registration_form.html'):
    """ Try to submit all the registration attributes for mojeID registration"""

    # Realm should be always something like 'https://example.org/openid/'
    realm = getattr(settings, 'MOJEID_REALM',
                    request.build_absolute_uri(reverse(top)))

    user = OpenIDBackend.get_user_from_request(request)
    user_id = user.pk if user else None

    # Create Nonce
    nonce = Nonce(server_url=realm, user_id=user_id,
                  timestamp=time.time(), salt=randomString(35, NONCE_CHARS))
    nonce.save()

    fields = []
    attributes = [x for x in get_attributes(attribute_set) if x.type == 'attribute']
    # Append attributes to creation request if user is valid
    if user:
        for attribute in attributes:
            form_attr = attribute.registration_form_attrs_html(user_id)
            if form_attr:
                fields.append(form_attr)

    # Render the redirection template
    return render_to_response(
        template_name,
        {
            'fields': fields,
            'action': get_registration_url(),
            'realm': realm,
            'nonce': nonce.registration_nonce,
        },
        context_instance=RequestContext(request)
    )


@csrf_exempt
def login_complete(request):
    # Get addres where to redirect after the login
    redirect_to = sanitise_redirect_url(
        request.session.get(mojeid_settings.MOJEID_SESSION_NEXT_PAGE_ATTR))
    attribute_set = request.session.get(SESSION_ATTR_SET_KEY, 'default')

    # clean the session
    if mojeid_settings.MOJEID_SESSION_NEXT_PAGE_ATTR in request.session:
        del request.session[mojeid_settings.MOJEID_SESSION_NEXT_PAGE_ATTR]

    if SESSION_ATTR_SET_KEY in request.session:
        del request.session[SESSION_ATTR_SET_KEY]

    # Get OpenID response and test whether it is valid

    endpoint = create_service()
    message = Message.fromPostArgs(request.REQUEST)
    consumer = MojeIDConsumer(DjangoOpenIDStore())

    try:
        openid_response = consumer.complete(
            message, endpoint, request.build_absolute_uri())
    except HTTPFetchingError:
        # if not using association and can't contact MojeID server
        return render_failure(request, errors.EndpointError())

    # Check whether the user is already logged in
    user_orig = OpenIDBackend.get_user_from_request(request)
    user_model = get_user_model()

    if openid_response.status == SUCCESS:

        try:
            if user_orig:
                # Send a signal to obtain HttpResponse
                resp = associate_user.send(
                    sender=__name__, request=request, openid_response=openid_response,
                    attribute_set=attribute_set, redirect=redirect_to
                )
                resp = [r[1] for r in resp if isinstance(r[1], HttpResponse)]
                if resp:
                    # Return first valid response
                    return resp[0]

                # Create association with currently logged in user
                OpenIDBackend.associate_openid_response(user_orig, openid_response)
            else:
                # Authenticate mojeID user.
                # Send a signal to obtain HttpResponse
                resp = authenticate_user.send(
                    sender=__name__, request=request, openid_response=openid_response,
                    attribute_set=attribute_set, redirect=redirect_to
                )
                resp = [r[1] for r in resp if isinstance(r[1], HttpResponse)]
                if resp:
                    # Return first valid response
                    return resp[0]

                # Perform a default action
                user_new = OpenIDBackend.authenticate_using_all_backends(
                    openid_response=openid_response, attribute_set=attribute_set)
                if not user_new:
                    # Failed to create a user
                    return render_failure(request, errors.UnknownUser())
                if not OpenIDBackend.is_user_active(user_new):
                    # user is deactivated
                    return render_failure(request, errors.DisabledAccount(user_new))
                # Create an association with the new user
                OpenIDBackend.associate_user_with_session(request, user_new)
        except DjangoOpenIDException as e:
            # Something went wrong
            user_id = None
            try:
                # Try to get user id
                user_id = UserOpenID.objects.get(claimed_id=openid_response.identity_url).user_id
            except (UserOpenID.DoesNotExist, user_model.DoesNotExist):
                # Report an error with identity_url
                user_login_report.send(
                    sender=__name__, request=request, username=openid_response.identity_url,
                    method='openid', success=False
                )

            # Report an error with the username
            user_login_report.send(
                sender=__name__, request=request, username=openid_response.identity_url,
                user_id=user_id, method='openid', success=False
            )

            # Render the failure page
            return render_failure(request, errors.AuthenticationFailed(e))

        response = HttpResponseRedirect(redirect_to)

        # Send signal to log the successful login attempt
        user_login_report.send(
            sender=__name__, request=request,
            user_id=user_orig.id if user_orig else user_new.id, method='openid', success=True
        )

        return response

    # Render other failures
    elif openid_response.status == FAILURE:
        user_login_report.send(
            sender=__name__, request=request, username=openid_response.identity_url,
            method='openid', success=False
        )
        return render_failure(request, errors.OpenIDAuthenticationFailed(openid_response))

    elif openid_response.status == CANCEL:
        user_login_report.send(
            sender=__name__, request=request, username=openid_response.identity_url,
            method='openid', success=False
        )
        return render_failure(request, errors.OpenIDAuthenticationCanceled())
    else:
        user_login_report.send(
            sender=__name__, request=request, username=openid_response.identity_url,
            method='openid', success=False
        )
        return render_failure(request, errors.OpenIDUnknownResponseType(openid_response))


@csrf_exempt
def assertion(request):
    """
    mojeID server connects here to propagate a response to the registration
    """
    def _reject(request, error):
        """ Reject response """
        return HttpResponse(dictToKV({'mode': 'reject', 'reason': error}))

    def _accept(request):
        """ Accept response """
        return HttpResponse(dictToKV({'mode': 'accept'}))

    # Accept only post
    if not request.method == 'POST':
        return _reject(request, Assertion.ErrorString.BAD_REQUEST)

    # Accept only valid status
    status = request.POST.get('status', None)
    if not status:
        return _reject(request, Assertion.ErrorString.MISSING_STATUS)
    if status not in Assertion.StatusCodes:
        return _reject(request, Assertion.ErrorString.INVALID_STATUS)

    # TODO check whether this request is from mojeID server and uses https with a proper certificate

    # Test calimed ID
    claimed_id = request.POST.get('claimed_id')
    if not claimed_id:
        return _reject(request, Assertion.ErrorString.MISSING_CLAIMED_ID)

    # The user was registered for mojeID
    if status == Assertion.StatusCodes.REGISTERED:
        registration_nonce = request.POST.get('registration_nonce')
        if registration_nonce is None:
            return _reject(request, Assertion.ErrorString.MISSING_NONCE)

        # check nonce
        try:
            nonce = Nonce.get_registration_nonce(registration_nonce)
        except Nonce.DoesNotExist:
            return _reject(request, Assertion.ErrorString.INVALID_NONCE)

        user_id = nonce.user_id
        nonce.delete()

        # Try to associate the user with mojeID
        if user_id:
            # Fetch the user
            user_model = get_user_model()
            try:
                user = user_model.objects.get(pk=user_id)
                # Create association
                OpenIDBackend.associate_openid(user, claimed_id)
            except (user_model.DoesNotExist, IdentityAlreadyClaimed):
                # Don't associte the user when the user doesn't exist or is already claimed
                # And assume that server sent us a valid claimed_id
                #
                # Note that user might been deleted before this assertion is triggered
                # Or the newly created mojeID account might been already associated
                # with a local account by the client
                #
                # Both of these cases are not considered as errors
                pass

    return _accept(request)


def top(request, template_name='openid/top.html'):
    """ The openid Endpoint
        this page should be only accessible by mojeID server
    """
    url = request.build_absolute_uri(reverse(xrds))
    title = getattr(settings, 'OPENID_APP_TITLE', 'OpenID Backend')
    return render_to_response(
        template_name,
        {'url': url, 'title': title},
        context_instance=RequestContext(request)
    )


def xrds(request, template_name='openid/xrds.xml'):
    """ Render xrds file
        This file should contain assertion and return_to urls.
    """
    return_to_url = request.build_absolute_uri(reverse(login_complete))
    assertion_url = request.build_absolute_uri(reverse(assertion))
    return render_to_response(
        template_name,
        {
            'return_to_url': return_to_url,
            'assertion_url': assertion_url
        },
        context_instance=RequestContext(request),
        content_type=YADIS_CONTENT_TYPE
    )


@require_POST
def disassociate(request):
    """
        Disassociate current user with OpenID
    """

    # Get the User
    user = OpenIDBackend.get_user_from_request(request)
    if not user:
        raise Http404

    # Get OpenID association
    association = OpenIDBackend.get_user_association(user)
    if not association:
        raise Http404

    # Remove the association
    association.delete()

    # Redirect back
    redirect = OpenIDBackend.get_redirect_to(request)
    redirect = redirect if redirect else getattr(settings, 'LOGIN_REDIRECT_URL', '/')
    return HttpResponseRedirect(sanitise_redirect_url(redirect))



from django.utils.translation import ugettext_lazy as _

_handlers = {}


class HandlerNotFound(Exception):
    pass


def register_handler(handler_name):

    def function_wrapper(f):

        def wrapped_function(*args, **kwargs):
            return f(*args, **kwargs)

        # This import is required to match the correct import
        from django_mojeid.attribute_handlers import _handlers

        # Store function into dict under handler name
        _handlers[handler_name] = wrapped_function

        return wrapped_function

    return function_wrapper


def call_handler(handler_name, *args, **kwargs):
    try:
        _handlers[handler_name](*args, **kwargs)
    except KeyError:
        raise HandlerNotFound(_('Handler with name %s was not found.') % handler_name)



# django-openid-auth -  OpenID integration for django.contrib.auth
#
# Copyright (C) 2007 Simon Willison
# Copyright (C) 2008-2013 Canonical Ltd.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from django.conf import settings
from django.conf.urls import patterns, url

from django_mojeid import views

urlpatterns = patterns(
    '',
    url(r'^$', views.top, name='openid-top'),
    url(r'^xrds.xml$', views.xrds, name='openid-xrds'),
    url(r'^disassociate/$', views.disassociate, name='openid-disassociate'),
    url(r'^initiate/$', views.login_begin, name='openid-init'),
    url(r'^initiate/(?P<attribute_set>\w*)$', views.login_begin, name='openid-init'),
    url(r'^complete/$', views.login_complete, name='openid-complete'),
)

if getattr(settings, 'USE_MOJEID_REGISTRATION_URLS', True):
    urlpatterns += patterns(
        '',
        url(r'^registration/$', views.registration, name='openid-registration'),
        url(r'^registration/(?P<attribute_set>\w*)$', views.registration,
            name='openid-registration'),
        url(r'^assertion/$', views.assertion, name='openid-assertion'),
    )



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured


# defaults

MOJEID_LOGIN_METHOD = "ANY"
MOJEID_INSTANCE_PRODUCTION = False
MOJEID_MAX_AUTH_AGE = None
MOJEID_SESSION_NEXT_PAGE_ATTR = 'mojeid_next_page'

class Settings(object):
    def __getattr__(self, name):
        try:
            attr = getattr(settings, name)
        except AttributeError:
            try:
                attr = globals()[name]
            except KeyError:
                raise AttributeError(
                    "'Settings' object has no attribute '%s'" % name)

        # validate
        if name == 'MOJEID_LOGIN_METHOD' and attr not in ("ANY", "CERT", "OTP"):
            raise ImproperlyConfigured(
                "Invalid MOJEID_LOGIN_METHOD '%s'" % attr)

        if name == 'MOJEID_MAX_AUTH_AGE' and not (
                attr is None or
                (isinstance(attr, int) and attr >= 0)):
            raise ImproperlyConfigured(
                "MOJEID_MAX_AUTH_AGE must be a positive integer (>= 0) or None")
        return attr

mojeid_settings = Settings()



# django-openid-auth -  OpenID integration for django.contrib.auth
#
# Copyright (C) 2007 Simon Willison
# Copyright (C) 2008-2013 Canonical Ltd.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from django.db.models.signals import post_delete
from django.dispatch import Signal, receiver
from django.contrib.auth import get_user_model


user_login_report = Signal(providing_args=[
    'request', 'username', 'user_id', 'method', 'success'])

trigger_error = Signal(providing_args=['request', 'error'])

authenticate_user = Signal(providing_args=['request', 'openid_response', 'redirect'])

associate_user = Signal(providing_args=['request', 'openid_response', 'redirect'])

# Fetch the delete user
user_model = get_user_model()


@receiver(post_delete, sender=user_model, dispatch_uid='user_delete')
def delete_association(**kwargs):

    from django_mojeid.models import UserOpenID

    sender = kwargs['sender']
    user = kwargs['instance']
    if sender == user_model:
        UserOpenID.objects.filter(user_id=user.pk).delete()



# django-openid-auth -  OpenID integration for django.contrib.auth
#
# Copyright (C) 2013 CZ.NIC
# Copyright (C) 2008-2013 Canonical Ltd.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""OpenID authentication"""

from __future__ import unicode_literals

from django.conf import settings
from django.core.exceptions import ValidationError
from django.contrib.auth import get_user_model
from django.utils.translation import ugettext_lazy as _

from openid.consumer.consumer import SUCCESS
from openid.extensions import ax

from django_mojeid.exceptions import (
    IdentityAlreadyClaimed,
    DuplicateUserViolation,
)
from django_mojeid.mojeid import get_attributes
from django_mojeid.attribute_handlers import call_handler


class OpenIDBackend:
    """A backend that authenticates the user based on an OpenID response."""

    supports_object_permissions = False
    supports_anonymous_user = True

    @classmethod
    def get_user_from_request(cls, request):
        """This method can be overwritten to implement custom user/session mechanizms
        currently it uses standard django.contrib.auth"""
        if not hasattr(request, 'user') or not hasattr(request.user, 'is_authenticated'):
            return None
        return request.user if request.user.is_authenticated() else None

    @classmethod
    def is_user_active(cls, user):
        """This method can be overwritten to implement custom user/session mechanizms
        currently it uses standard django.contrib.auth"""
        return user.is_active if user else False

    @classmethod
    def is_user_authenticated(cls, user):
        """This method can be overwritten to implement custom user/session mechanizms
        currently it uses standard django.contrib.auth"""
        return user.is_authenticated() if user else False

    @classmethod
    def is_user_associated_with_openid(cls, user):
        if not user:
            return False
        from django_mojeid.models import UserOpenID
        return UserOpenID.objects.filter(user_id=user.pk).exists()

    @classmethod
    def get_user_association(cls, user):
        from django_mojeid.models import UserOpenID
        if not user:
            return None
        try:
            association = UserOpenID.objects.get(user_id=user.pk)
        except UserOpenID.DoesNotExist:
            association = None

        return association

    @classmethod
    def get_redirect_to(cls, request):
        """This method can be overwritten to implement custom user/session mechanizms
        currently it uses standard django.contrib.auth"""
        return request.REQUEST.get(cls.get_redirect_field_name(), '')

    @classmethod
    def get_redirect_field_name(cls):
        """This method can be overwritten to implement custom user/session mechanizms
        currently it uses standard django.contrib.auth"""
        from django.contrib.auth import REDIRECT_FIELD_NAME
        return REDIRECT_FIELD_NAME

    @staticmethod
    def authenticate_using_all_backends(**credentials):
        """This method can be overwritten to implement custom user/session mechanizms
        currently it uses standard django.contrib.auth"""
        from django.contrib.auth import authenticate
        return authenticate(**credentials)

    @classmethod
    def associate_user_with_session(cls, request, user):
        """This method can be overwritten to implement custom user/session mechanizms
        currently it uses standard django.contrib.auth"""
        from django.contrib.auth import login as auth_login

        # skip when the user is not authenticated (=AnonymousUser)
        if not user or not user.is_authenticated():
            return

        # Set backend if it is not set
        if not hasattr(user, 'backend'):
            setattr(user, 'backend', 'django_mojeid.auth.OpenIDBackend')

        auth_login(request, user)

    def authenticate(self, **kwargs):
        """Authenticate the user based on an OpenID response."""
        # Require that the OpenID response be passed in as a keyword
        # argument, to make sure we don't match the username/password
        # calling conventions of authenticate.

        from django_mojeid.models import UserOpenID

        openid_response = kwargs.get('openid_response')
        if openid_response is None:
            return None

        attribute_set = kwargs.get('attribute_set', 'default')

        if openid_response.status != SUCCESS:
            return None

        user = None
        new_user = False
        try:
            user_openid = UserOpenID.objects.get(
                claimed_id__exact=openid_response.identity_url)
        except UserOpenID.DoesNotExist:
            if getattr(settings, 'OPENID_CREATE_USERS', False):
                user = self.create_user_from_openid(openid_response, attribute_set)
                new_user = True
        else:
            user_model = get_user_model()
            try:
                user = user_model.objects.get(pk=user_openid.user_id)
            except user_model.DoesNotExist:
                user = None

        if user is None:
            return None

        if not new_user:
            self.update_user_from_openid(user.pk, openid_response, attribute_set)

        # Run custom Attribute handler
        OpenIDBackend.run_handlers(openid_response, user, attribute_set)

        return user

    @staticmethod
    def get_model_changes(openid_response, only_updatable=False,
                          attribute_set='default'):

        attributes = [x for x in get_attributes(attribute_set)
                      if x.type in ['attribute', 'internal']]

        fetch_response = ax.FetchResponse.fromSuccessResponse(openid_response)

        res = {}

        # filter remove non-updatable attributes
        if only_updatable:
            attributes = [x for x in attributes if x.updatable]

        for attribute in attributes:
            if attribute.model not in res:
                res[attribute.model] = {'user_id_field_name': attribute.user_id_field_name}
            val = attribute.get_value(fetch_response, attribute.required,
                                      openid_response=openid_response)

            if val is not None:
                res[attribute.model][attribute.modelAttribute] = val

        return res

    @staticmethod
    def run_handlers(openid_response, user, attribute_set='default'):
        handlers = [x for x in get_attributes(attribute_set) if x.type == 'handler']

        if not handlers:
            return

        fetch_response = ax.FetchResponse.fromSuccessResponse(openid_response)

        for handler in handlers:
            val = handler.attribute.get_value(fetch_response, handler.required, openid_response)
            call_handler(handler.name, user, val)

    def create_user_from_openid(self, openid_response, attribute_set='default'):
        changes = OpenIDBackend.get_model_changes(openid_response, attribute_set=attribute_set)

        user_model = get_user_model()

        # Id will be generated no need to set this field
        del changes[user_model]['user_id_field_name']

        # Create the main user structure
        user = user_model(**changes[user_model])
        try:
            user.validate_unique()
        except ValidationError as e:
            raise DuplicateUserViolation(", ".join(e.messages))
        user.save()

        # User created remove it from the dict
        del changes[user_model]

        # Create other structures
        for model, kwargs in changes.items():
            foreign_key_name = kwargs['user_id_field_name']
            del kwargs['user_id_field_name']
            kwargs[foreign_key_name] = user.pk
            m = model(**kwargs)
            m.save()

        OpenIDBackend.associate_openid_response(user, openid_response)

        return user

    @classmethod
    def update_user_from_openid(cls, user_id, openid_response, attribute_set='default'):
        changes = OpenIDBackend.get_model_changes(openid_response, only_updatable=True,
                                                  attribute_set=attribute_set)

        for model, kwargs in changes.items():
            foreign_key_name = kwargs['user_id_field_name']
            del kwargs['user_id_field_name']
            model.objects.filter(**{foreign_key_name: user_id}).update(**kwargs)

        return changes

    @staticmethod
    def associate_openid_response(user, openid_response):
        """Associate an OpenID request with a user account."""
        # Check to see if this OpenID has already been claimed.
        return OpenIDBackend.associate_openid(user, openid_response.identity_url)

    @staticmethod
    def associate_openid(user, claimed_id):
        """Associate an OpenID with a user account."""

        from django_mojeid.models import UserOpenID

        # Check to see if this OpenID has already been claimed.
        try:
            user_openid = UserOpenID.objects.get(
                claimed_id__exact=claimed_id)
        except UserOpenID.DoesNotExist:
            user_openid = UserOpenID(
                user_id=user.pk,
                claimed_id=claimed_id)
            user_openid.save()
        else:
            if user_openid.user_id != user.pk:
                raise IdentityAlreadyClaimed(
                    _("The identity %s has already been claimed")
                    % claimed_id)

        return user_openid

    @classmethod
    def get_user(cls, user_id):
        try:
            user_model = get_user_model()
            return user_model.objects.get(pk=user_id)
        except user_model.DoesNotExist:
            return None



# django-openid-auth -  OpenID integration for django.contrib.auth
#
# Copyright (C) 2013 CZ.NIC
# Copyright (C) 2008-2013 Canonical Ltd.
# Copyright (C) 2010 Dave Walker
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from django.conf import settings
from django.contrib import admin
from django.utils.translation import ugettext_lazy as _

from django_mojeid.models import Nonce, Association, UserOpenID
from django_mojeid.store import DjangoOpenIDStore


class NonceAdmin(admin.ModelAdmin):
    list_display = ('server_url', 'timestamp')
    actions = ['cleanup_nonces']

    def cleanup_nonces(self, request, queryset):
        store = DjangoOpenIDStore()
        count = store.cleanupNonces()
        self.message_user(request, _("%d expired nonces removed") % count)
    cleanup_nonces.short_description = _("Clean up expired nonces")

admin.site.register(Nonce, NonceAdmin)


class AssociationAdmin(admin.ModelAdmin):
    list_display = ('server_url', 'assoc_type')
    list_filter = ('assoc_type',)
    search_fields = ('server_url',)
    actions = ['cleanup_associations']

    def cleanup_associations(self, request, queryset):
        store = DjangoOpenIDStore()
        count = store.cleanupAssociations()
        self.message_user(request, _("%d expired associations removed") % count)
    cleanup_associations.short_description = _("Clean up expired associations")

admin.site.register(Association, AssociationAdmin)


class UserOpenIDAdmin(admin.ModelAdmin):
    list_display = ('user_id', 'claimed_id')
    search_fields = ('claimed_id',)

admin.site.register(UserOpenID, UserOpenIDAdmin)


# Support for allowing openid authentication for /admin (django.contrib.admin)
if getattr(settings, 'OPENID_USE_AS_ADMIN_LOGIN', False):
    from django.http import HttpResponseRedirect
    from django_mojeid import views

    def _openid_login(self, request, error_message='', extra_context=None):
        if request.user.is_authenticated():
            if not request.user.is_staff:
                return views.default_render_failure(
                    request, _("User %s does not have admin access.")
                    % request.user.username)
            assert error_message, _("Unknown Error: %s") % error_message
        else:
            # Redirect to openid login path,
            return HttpResponseRedirect(
                settings.LOGIN_URL + "?next=" + request.get_full_path())

    # Overide the standard admin login form.
    admin.sites.AdminSite.login = _openid_login



from __future__ import unicode_literals

import sys

from django.utils.translation import ugettext_lazy as _


if sys.version_info.major != 2:
    unicode = str


class GeneralError(object):
    http_status = 500
    name = 'general'
    text = _('General OpenID error.')

    @property
    def msg(self):
        return self.text


class DiscoveryError(GeneralError):
    http_status = 404
    name = 'discovery'
    text = _("OpenID discovery error")

    def __init__(self, exception):
        self.exception = exception

    @property
    def msg(self):
        return '%s: %s' % (self.text, unicode(self.exception))


class EndpointError(GeneralError):
    name = 'endpoint'
    text = _('This is an OpenID relying party endpoint.')


class UnknownUser(GeneralError):
    http_status = 404
    name = 'unknown_user'
    text = _('Unknown User')


class DisabledAccount(GeneralError):
    http_status = 403
    name = 'disabled_account'
    text = _('Disabled Account')

    def __init__(self, user):
        self.user = user

    @property
    def msg(self):
        return '%s for User with id=%d' % (self.text, self.user.pk)


class AuthenticationFailed(GeneralError):
    http_status = 403
    name = 'auth_failed'
    text = _('Authentication Failed')

    def __init__(self, exception):
        self.exception = exception

    @property
    def msg(self):

        return '%s: %s' % (self.text, unicode(self.exception))


class OpenIDAuthenticationFailed(GeneralError):
    http_status = 403
    name = 'openid_auth_failed'
    text = _('OpenID authentication failed')

    def __init__(self, openid_response):
        self.openid_response = openid_response

    @property
    def msg(self):
        return '%s: %s' % (self.text, unicode(self.openid_response.message))


class OpenIDAuthenticationCanceled(GeneralError):
    http_status = 403
    name = 'openid_auth_cancelled'
    text = _('OpenID Authentication Cancelled')


class OpenIDUnknownResponseType(GeneralError):
    name = 'openid_unknown'
    text = _('OpenID Unknown Response Type')

    def __init__(self, openid_response):
        self.openid_response = openid_response

    @property
    def msg(self):
        return '%s: %d' % (self.text, self.openid_response.status)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# django-mojeid - mojeID integration for django
#
# Copyright (C) 2013-2015 CZ.NIC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import unicode_literals

from urlparse import urlparse

from django.core.exceptions import FieldError, ImproperlyConfigured
from django.http import Http404
from django.utils.translation import ugettext_lazy as _
from django.utils.translation import ugettext

from openid.consumer.consumer import GenericConsumer, FailureResponse, SUCCESS
from openid.consumer.discover import OpenIDServiceEndpoint
from openid.extensions import ax, pape

from django_mojeid.exceptions import RequiredAttributeNotReturned
from django_mojeid.settings import mojeid_settings


mojeid_services = {
    'testing': {
        'url': "https://mojeid.fred.nic.cz/endpoint/",
        'registration': 'https://mojeid.fred.nic.cz/registration/endpoint/',
        'xrds': """<?xml version="1.0" encoding="UTF-8"?>
<xrds:XRDS xmlns="xri://$xrd*($v*2.0)" xmlns:xrds="xri://$xrds">
<XRD>
<Service>
    <Type>http://specs.openid.net/auth/2.0/server</Type>
    <Type>http://specs.openid.net/extensions/pape/1.0</Type>
    <Type>http://openid.net/extensions/sreg/1.1</Type>
    <Type>http://openid.net/srv/ax/1.0</Type>
    <URI>https://mojeid.fred.nic.cz/endpoint/</URI>
</Service>
</XRD>
</xrds:XRDS>
"""
    },
    'production': {
        'url': "https://mojeid.cz/endpoint/",
        'registration': 'https://mojeid.cz/registration/endpoint/',
        'xrds': """<?xml version="1.0" encoding="UTF-8"?>
<xrds:XRDS xmlns="xri://$xrd*($v*2.0)" xmlns:xrds="xri://$xrds">
<XRD>
<Service>
    <Type>http://specs.openid.net/auth/2.0/server</Type>
    <Type>http://specs.openid.net/extensions/pape/1.0</Type>
    <Type>http://openid.net/extensions/sreg/1.1</Type>
    <Type>http://openid.net/srv/ax/1.0</Type>
    <URI>https://mojeid.cz/endpoint/</URI>
</Service>
</XRD>
</xrds:XRDS>
"""
    }
}


def create_service():
    endpoint_type = 'production' if mojeid_settings.MOJEID_INSTANCE_PRODUCTION \
                    else 'testing'
    defs = mojeid_services[endpoint_type]
    return OpenIDServiceEndpoint.fromXRDS(defs['url'], defs['xrds'])[0]


def get_registration_url():
    endpoint_type = 'production' if mojeid_settings.MOJEID_INSTANCE_PRODUCTION \
                    else 'testing'
    return mojeid_services[endpoint_type]['registration']


class MojeIDConsumer(GenericConsumer):
    def _verifyDiscoverySingle(self, endpoint, to_match):
        """This function normally verifies that the result (obtained from
        the mojeid server matches with xrds used in the beginning but as this
        two are the same, we can skip this check (and thus avoid downloading
        another xrds document from $username.mojeid.cz)
        """
        pass

    def complete(self, message, endpoint, return_to):
        response = super(MojeIDConsumer, self).complete(
            message, endpoint, return_to)

        if response.status != SUCCESS:
            return response

        # check if pape login method is the required one
        required_auth = {
            "OTP": pape.AUTH_MULTI_FACTOR,
            "CERT": pape.AUTH_PHISHING_RESISTANT
        }.get(mojeid_settings.MOJEID_LOGIN_METHOD, None)

        if required_auth:
            pape_response = pape.Response.fromSuccessResponse(response)
            if required_auth not in pape_response.auth_policies:
                return FailureResponse(endpoint, "Required authentication "
                                       "method was not used.")

        # check if all required attributes are present
        # TODO

        return response


def get_attributes(attribute_set):

    default = getattr(mojeid_settings, 'MOJEID_ATTRIBUTES', [])
    res = getattr(mojeid_settings, 'MOJEID_ATTRIBUTES_SETS', {})

    # MOJEID_ATTRIBUTES are default when present
    if default or not res:
        res['default'] = default

    try:
        return res[attribute_set]
    except KeyError:
        raise Http404


def get_attribute_query(attribute_set='defualt'):
    """ Return attributes without duplicities """
    attributes = get_attributes(attribute_set)

    used_dict = {}
    filtered_attributes = []
    for attribute in attributes:

        # Skip for the Internal attributes
        if attribute.type == 'internal':
            continue

        required = attribute.required
        attribute = attribute.attribute if attribute.type == 'handler' else attribute

        if attribute.code in used_dict:
            if not filtered_attributes[used_dict[attribute.code]][1]:
                # required=true has higher priority
                filtered_attributes[used_dict[attribute.code]] = (attribute, required, )
        else:
            used_dict[attribute.code] = len(filtered_attributes)
            filtered_attributes.append((attribute, required, ))

    return filtered_attributes


class CustomHandler(object):
    type = 'handler'

    def __init__(self, attribute, name, required=True):
        self.name = name
        self.required = required
        self.attribute = attribute


class Attribute(object):
    type = 'abstract'

    text = None

    def __init__(self, modelApp, modelClass, modelAttribute,
                 user_id_field_name='user_id', required=True, updatable=False):
        self.modelClass = modelClass
        self.modelApp = modelApp
        self.modelAttribute = modelAttribute
        self.user_id_field_name = user_id_field_name
        self.required = required
        self.updatable = updatable
        self._model = None

    @property
    def model(self):
        if not self._model:
            # Import model during runtime
            from django.db.models.loading import get_model
            # set the model
            self._model = get_model(self.modelApp, self.modelClass)
            if not self._model:
                raise ImproperlyConfigured(
                    _("Model '%(model)s' from App '%(app)s' does not exist.") %
                    {'model': self.modelClass, 'app': self.modelApp})
        return self._model

    # This method could be overwritten using inheritance
    def _set(self, record, attribute, value):
        # Simply just set the attribute
        setattr(record, attribute, value)

    # This method could be overwritten using inheritance
    def _get_record(self, id):
        packed = {self.user_id_field_name: id}
        return self.model.objects.get(**packed)

    # This needs to be overwritten using inheritance
    @classmethod
    def _get_value(cls, response):
        return None

    def set_model_value(self, id, value):
        record = self._get_record(id)
        if not hasattr(record, self.modelAttribute):
            raise FieldError(
                _("Cannot resolve keyword '%(model)s' into field. "
                  "Choices are: %(choices)s") %
                {"model": self.modelAttribute,
                 "choices": ", ".join(record._meta._name_map.keys())})
        self._set(record, self.modelAttribute, value)
        record.save()

    # This method could be overwritten using inheritance
    def _get_model_value(self, id):
        return getattr(self._get_record(id), self.modelAttribute)


class MojeIDAttribute(Attribute):
    type = 'attribute'

    code = None
    schema = None

    def __init__(self, modelApp, modelClass, modelAttribute,
                 user_id_field_name='user_id', required=True,
                 updatable=False, use_for_registration=True):
        self.use_for_registration = use_for_registration
        super(MojeIDAttribute, self).__init__(
            modelApp, modelClass, modelAttribute, user_id_field_name,
            required, updatable
        )

    @classmethod
    def _get_value(cls, response):
        return response.getSingle(cls.schema, None)

    @classmethod
    def generate_ax_attrinfo(cls, required):
        return ax.AttrInfo(cls.schema, alias=cls.code, required=required)

    @classmethod
    def get_value(cls, response, required, openid_response=None):
        value = cls._get_value(response)
        if required and value is None:
            raise RequiredAttributeNotReturned(
                ugettext("Required Attribute '%(code)s' (%(text)s) was not returned.")
                % {"code": cls.code, "text": cls.text}
            )
        return value

    def registration_form_attrs(self, id):
        # Return none if registration field is not present
        if not hasattr(self, 'registration_field') or not self.use_for_registration:
            return None

        value = self._get_model_value(id)
        # No value present
        if value is None:
            return None

        return {'name': self.registration_field, 'label': self.text, 'value': value}

    # This method could be overwritten using inheritance
    def _get_form_html_template(self):
        return '<label for="%s">%s</label><input type="text" name="%s" value="%s">'

    def registration_form_attrs_html(self, id):

        field = self.registration_form_attrs(id)

        # Field was not obtained.
        if not field:
            return ""

        return self._get_form_html_template() % (
            field['name'], field['label'], field['name'], field['value']
        )


class MojeIDBooleanAttribute(MojeIDAttribute):

    @classmethod
    def _get_value(self, response):
        res = response.getSingle(self.schema, None)
        if res is None:
            return None
        return True if res.lower() == 'true' else False


class BirthDate(MojeIDAttribute):
    code = 'birthdate'
    schema = 'http://axschema.org/birthDate'
    text = _('Birth Date')  # 'Datum Narození'
    registration_field = 'birth_date'


class FullName(MojeIDAttribute):
    code = 'fullname'
    schema = 'http://axschema.org/namePerson'
    text = _('Full Name')  # 'Celé jméno'


class FirstName(MojeIDAttribute):
    code = 'firstname'
    schema = 'http://axschema.org/namePerson/first'
    text = _('First Name')  # 'Jméno'
    registration_field = 'first_name'


class LastName(MojeIDAttribute):
    code = 'lastname'
    schema = 'http://axschema.org/namePerson/last'
    text = _('Last Name')  # 'Příjmení'
    registration_field = 'last_name'


class NickName(MojeIDAttribute):
    code = 'nick'
    schema = 'http://axschema.org/namePerson/friendly'
    text = _('Nick Name')  # 'Přezdívka'
    registration_field = 'username'


class Company(MojeIDAttribute):
    code = 'company'
    schema = 'http://axschema.org/company/name'
    text = _('Company')  # 'Jméno společnosti'
    registration_field = 'organization'


class HomeAddress(MojeIDAttribute):
    code = 'h_address'
    schema = 'http://axschema.org/contact/postalAddress/home'
    text = _('Home Address – Street')  # 'Domácí adresa – Ulice'
    registration_field = 'address__default__street1'


class HomeAddress2(MojeIDAttribute):
    code = 'h_address2'
    schema = 'http://axschema.org/contact/postalAddressAdditional/home'
    text = _('Home Address – Street2')  # 'Domácí adresa – Ulice2'
    registration_field = 'address__default__street2'


class HomeAddress3(MojeIDAttribute):
    code = 'h_address3'
    schema = 'http://specs.nic.cz/attr/addr/main/street3'
    text = _('Home Address – Street3')  # 'Domácí adresa – Ulice3'
    registration_field = 'address__default__street3'


class HomeCity(MojeIDAttribute):
    code = 'h_city'
    schema = 'http://axschema.org/contact/city/home'
    text = _('Home Address – City')  # 'Domácí adresa – Město'
    registration_field = 'address__default__city'


class HomeState(MojeIDAttribute):
    code = 'h_state'
    schema = 'http://axschema.org/contact/state/home'
    text = _('Home Address – State')  # 'Domácí adresa – Stát'
    registration_field = 'address__default__state'


class HomeCountry(MojeIDAttribute):
    code = 'h_country'
    schema = 'http://axschema.org/contact/country/home'
    text = _('Home Address – Country')  # 'Domácí adresa – Země'
    registration_field = 'address__default__country'


class HomePostCode(MojeIDAttribute):
    code = 'h_postcode'
    schema = 'http://axschema.org/contact/postalCode/home'
    text = _('Home Address – Country')  # 'Domácí adresa – PSČ'
    registration_field = 'address__default__postal_code'


class BillingAddress(MojeIDAttribute):
    code = 'b_address'
    schema = 'http://specs.nic.cz/attr/addr/bill/street'
    text = _('Billing Address – Street')  # 'Faktur. adresa – Ulice'
    registration_field = 'address__billing__street1'


class BillingAddress2(MojeIDAttribute):
    code = 'b_address2'
    schema = 'http://specs.nic.cz/attr/addr/bill/street2'
    text = _('Billing Address – Street2')  # 'Faktur. adresa – Ulice2'
    registration_field = 'address__billing__street2'


class BillingAddress3(MojeIDAttribute):
    code = 'b_address3'
    schema = 'http://specs.nic.cz/attr/addr/bill/street3'
    text = _('Billing Address – Street3')  # 'Faktur. adresa – Ulice3'
    registration_field = 'address__billing__street3'


class BillingCity(MojeIDAttribute):
    code = 'b_city'
    schema = 'http://specs.nic.cz/attr/addr/bill/city'
    text = _('Billing Address – City')  # 'Faktur. adresa – Město'
    registration_field = 'address__billing__city'


class BillingState(MojeIDAttribute):
    code = 'b_state'
    schema = 'http://specs.nic.cz/attr/addr/bill/sp'
    text = _('Billing Address – State')  # 'Faktur. adresa – Stát'
    registration_field = 'address__billing__state'


class BillingCountry(MojeIDAttribute):
    code = 'b_country'
    schema = 'http://specs.nic.cz/attr/addr/bill/cc'
    text = _('Billing Address – Country')  # 'Faktur. adresa – Země'
    registration_field = 'address__billing__country'


class BillingPostCode(MojeIDAttribute):
    code = 'b_postcode'
    schema = 'http://specs.nic.cz/attr/addr/bill/pc'
    text = _('Billing Address – Postal Code')  # 'Faktur. adresa – PSČ'
    registration_field = 'address__billing__postal_code'


class ShippingAddress(MojeIDAttribute):
    code = 's_address'
    schema = 'http://specs.nic.cz/attr/addr/ship/street'
    text = _('Shipping Address – Street')  # 'Doruč. adresa – Ulice'
    registration_field = 'address__shipping__street1'


class ShippingAddress2(MojeIDAttribute):
    code = 's_address2'
    schema = 'http://specs.nic.cz/attr/addr/ship/street2'
    text = _('Shipping Address – Street2')  # 'Doruč. adresa – Ulice2'
    registration_field = 'address__shipping__street2'


class ShippingAddress3(MojeIDAttribute):
    code = 's_address3'
    schema = 'http://specs.nic.cz/attr/addr/ship/street3'
    text = _('Shipping Address – Street3')  # 'Doruč. adresa – Ulice3'
    registration_field = 'address__shipping__street3'


class ShippingCity(MojeIDAttribute):
    code = 's_city'
    schema = 'http://specs.nic.cz/attr/addr/ship/city'
    text = _('Shipping Address – City')  # 'Doruč. adresa – Město'
    registration_field = 'address__shipping__city'


class ShippingState(MojeIDAttribute):
    code = 's_state'
    schema = 'http://specs.nic.cz/attr/addr/ship/sp'
    text = _('Shipping Address – State')  # 'Doruč. adresa – Stát'
    registration_field = 'address__shipping__state'


class ShippingCountry(MojeIDAttribute):
    code = 's_country'
    schema = 'http://specs.nic.cz/attr/addr/ship/cc'
    text = _('Shipping Address – Country')  # 'Doruč. adresa – Země'
    registration_field = 'address__shipping__country'


class ShippingPostCode(MojeIDAttribute):
    code = 's_postcode'
    schema = 'http://specs.nic.cz/attr/addr/ship/pc'
    text = _('Shipping Address – Postal Code')  # 'Doruč. adresa – PSČ'
    registration_field = 'address__shipping__postal_code'


class MailingAddress(MojeIDAttribute):
    code = 'm_address'
    schema = 'http://specs.nic.cz/attr/addr/mail/street'
    text = _('Mailing Address – Street')  # 'Koresp. adresa – Ulice'
    registration_field = 'address__mailing__street1'


class MailingAddress2(MojeIDAttribute):
    code = 'm_address2'
    schema = 'http://specs.nic.cz/attr/addr/mail/street2'
    text = _('Mailing Address – Street2')  # 'Koresp. adresa – Ulice2'
    registration_field = 'address__mailing__street2'


class MailingAddress3(MojeIDAttribute):
    code = 'm_address3'
    schema = 'http://specs.nic.cz/attr/addr/mail/street3'
    text = _('Mailing Address – Street3')  # 'Koresp. adresa – Ulice3'
    registration_field = 'address__mailing__street3'


class MailingCity(MojeIDAttribute):
    code = 'm_city'
    schema = 'http://specs.nic.cz/attr/addr/mail/city'
    text = _('Mailing Address – City')  # 'Koresp. adresa – Město'
    registration_field = 'address__mailing__city'


class MailingState(MojeIDAttribute):
    code = 'm_state'
    schema = 'http://specs.nic.cz/attr/addr/mail/sp'
    text = _('Mailing Address – State')  # 'Koresp. adresa – Stát'
    registration_field = 'address__mailing__state'


class MailingCountry(MojeIDAttribute):
    code = 'm_country'
    schema = 'http://specs.nic.cz/attr/addr/mail/cc'
    text = _('Mailing Address – Country')  # 'Koresp. adresa – Země'
    registration_field = 'address__mailing__country'


class MailingPostCode(MojeIDAttribute):
    code = 'm_postcode'
    schema = 'http://specs.nic.cz/attr/addr/mail/pc'
    text = _('Mailing Address – Postal Code')  # 'Koresp. adresa – PSČ'
    registration_field = 'address__mailing__postal_code'


class Phone(MojeIDAttribute):
    code = 'phone'
    schema = 'http://axschema.org/contact/phone/default'
    text = _('Phone – Default')  # 'Telefon – Hlavní'
    registration_field = 'phone__default__number'


class PhoneHome(MojeIDAttribute):
    code = 'phone_home'
    schema = 'http://axschema.org/contact/phone/home'
    text = _('Phone – Home')  # 'Telefon – Domácí'
    registration_field = 'phone__home__number'


class PhoneWork(MojeIDAttribute):
    code = 'phone_work'
    schema = 'http://axschema.org/contact/phone/business'
    text = _('Phone – Work')  # 'Telefon – Pracovní'
    registration_field = 'phone__office__number'


class PhoneMobile(MojeIDAttribute):
    code = 'phone_mobile'
    schema = 'http://axschema.org/contact/phone/cell'
    text = _('Phone – Mobile')  # 'Telefon – Mobil'
    registration_field = 'phone__mobile__number'


class Fax(MojeIDAttribute):
    code = 'fax'
    schema = 'http://axschema.org/contact/phone/fax'
    text = _('Fax')  # 'Telefon – Fax'


class Email(MojeIDAttribute):
    code = 'email'
    schema = 'http://axschema.org/contact/email'
    text = _('Email – Default')  # 'Email – Hlavní'
    registration_field = 'email__default__email'


class Email2(MojeIDAttribute):
    code = 'email2'
    schema = 'http://specs.nic.cz/attr/email/notify'
    text = _('Email – Notify')  # 'Email – Notifikační'
    registration_field = 'email__notify__email'


class Email3(MojeIDAttribute):
    code = 'email3'
    schema = 'http://specs.nic.cz/attr/email/next'
    text = _('Email – Other')  # 'Email – Další'
    registration_field = 'email__next__email'


class Url(MojeIDAttribute):
    code = 'url'
    schema = 'http://axschema.org/contact/web/default'
    text = _('URL – Default')  # 'URL – Hlavní'
    registration_field = 'urladdress__main__url'


class Blog(MojeIDAttribute):
    code = 'blog'
    schema = 'http://axschema.org/contact/web/blog'
    text = _('URL – Blog')  # 'URL – Blog'
    registration_field = 'urladdress__blog__url'


class Url2(MojeIDAttribute):
    code = 'url2'
    schema = 'http://specs.nic.cz/attr/url/personal'
    text = _('URL – Personal')  # 'URL – Osobní'
    registration_field = 'urladdress__personal__url'


class Url3(MojeIDAttribute):
    code = 'url3'
    schema = 'http://specs.nic.cz/attr/url/work'
    text = _('URL – Work')  # 'URL – Pracovní'
    registration_field = 'urladdress__office__url'


class RSS(MojeIDAttribute):
    code = 'rss'
    schema = 'http://specs.nic.cz/attr/url/rss'
    text = _('URL – RSS')  # 'URL – RSS'
    registration_field = 'urladdress__rss__url'


class Facebook(MojeIDAttribute):
    code = 'fb'
    schema = 'http://specs.nic.cz/attr/url/facebook'
    text = _('URL – Facebook')  # 'URL – Facebook'
    registration_field = 'urladdress__facebook__url'


class Twitter(MojeIDAttribute):
    code = 'twitter'
    schema = 'http://specs.nic.cz/attr/url/twitter'
    text = _('URL – Twitter')  # 'URL – Twitter'
    registration_field = 'urladdress__twitter__url'


class LinkedIn(MojeIDAttribute):
    code = 'linkedin'
    schema = 'http://specs.nic.cz/attr/url/linkedin'
    text = _('URL – LinkedIN')  # 'URL – LinkedIN'
    registration_field = 'urladdress__linkedin__url'


class ICQ(MojeIDAttribute):
    code = 'icq'
    schema = 'http://axschema.org/contact/IM/ICQ'
    text = _('IM – ICQ')  # 'IM – ICQ'
    registration_field = 'imaccount__icq__username'


class Jabber(MojeIDAttribute):
    code = 'jabber'
    schema = 'http://axschema.org/contact/IM/Jabber'
    text = _('IM – Jabber')  # 'IM – Jabber'
    registration_field = 'imaccount__jabber__username'


class Skype(MojeIDAttribute):
    code = 'skype'
    schema = 'http://axschema.org/contact/IM/Skype'
    text = _('IM – Skype')  # 'IM – Skype'
    registration_field = 'imaccount__skype__username'


class GoogleTalk(MojeIDAttribute):
    code = 'gtalk'
    schema = 'http://specs.nic.cz/attr/im/google_talk'
    text = _('IM – Google Talk')  # 'IM – Google Talk'
    registration_field = 'imaccount__google_talk__username'


class WindowsLive(MojeIDAttribute):
    code = 'wlive'
    schema = 'http://specs.nic.cz/attr/im/windows_live'
    text = _('IM – Windows Live')  # 'IM – Windows Live'
    registration_field = 'imaccout__windows_live__username'


class ICO(MojeIDAttribute):
    code = 'vat_id'
    schema = 'http://specs.nic.cz/attr/contact/ident/vat_id'
    text = _('Identifier – VAT Identification Number')  # 'Identifikátor – ICO'
    registration_field = 'vat_id_num'


class DIC(MojeIDAttribute):
    code = 'vat'
    schema = 'http://specs.nic.cz/attr/contact/vat'
    text = _('Identifier – VAT registration Number')  # 'Identifikátor – DIC'
    registration_field = 'vat_reg_num'


class IdentityCard(MojeIDAttribute):
    code = 'op'
    schema = 'http://specs.nic.cz/attr/contact/ident/card'
    text = _('Identifier – ID Card')  # 'Identifikátor – OP'
    registration_field = 'id_card_num'


class Passport(MojeIDAttribute):
    code = 'pas'
    schema = 'http://specs.nic.cz/attr/contact/ident/pass'
    text = _('Identifier – Passport')  # 'Identifikátor – PAS'
    registration_field = 'passport_num'


class MPSV(MojeIDAttribute):
    code = 'mpsv'
    schema = 'http://specs.nic.cz/attr/contact/ident/ssn'
    text = _('Identifier – MPSV')  # 'Identifikátor – MPSV'
    registration_field = 'ssn_id_num'


class Student(MojeIDBooleanAttribute):
    code = 'student'
    schema = 'http://specs.nic.cz/attr/contact/student'
    text = _('Flag – Student')  # 'Příznak – Student'


class Validated(MojeIDBooleanAttribute):
    # Probably not supported anymore
    code = 'validated'
    schema = 'http://specs.nic.cz/attr/contact/valid'
    text = _('Flag – Validation')  # 'Příznak – Validace'


class Status(MojeIDAttribute):
    # Probably not supported anymore
    code = 'status'
    schema = 'http://specs.nic.cz/attr/contact/status'
    text = _('Account State')  # 'Stav účtu'


class Adult(MojeIDBooleanAttribute):
    code = 'adult'
    schema = 'http://specs.nic.cz/attr/contact/adult'
    text = _('Flag – Adult')  # 'Příznak – Starší 18 let'


class Image(MojeIDAttribute):
    code = 'image'
    schema = 'http://specs.nic.cz/attr/contact/image'
    text = _('Image (base64)')  # 'Obrázek (base64)'


class InteralAttribute(Attribute):
    type = 'internal'

    @classmethod
    def get_value(cls, response, required, openid_response):
        value = cls._get_value(openid_response)
        if required and value is None:
            raise RequiredAttributeNotReturned(
                ugettext("Required Internal Attribute '%(code)s' (%(text)s) was not returned.")
                % {"code": unicode(cls.code), "text": unicode(cls.text)}
            )
        return value


class DisplayID(InteralAttribute):
    text = _(u'ID to display')

    @classmethod
    def _get_value(cls, response):
        return urlparse(response.getDisplayIdentifier()).netloc


class LoginID(InteralAttribute):
    text = _(u'mojeID login')

    @classmethod
    def _get_value(cls, response):
        return str(urlparse(response.getDisplayIdentifier()).netloc).split(".", 1)[0]


class Assertion:

    # Status codes for the assertions
    class StatusCodes:
        _codes = [
            'REGISTERED',
            'CONDITIONALLY_IDENTIFIED',
            'IDENTIFIED',
            'VALIDATED',
        ]

        class __metaclass__(type):
            def __iter__(self):
                for attr in self._codes:
                    yield attr

            def __getattr__(self, name):
                if name not in self._codes:
                    raise AttributeError(
                        _("class %(class)s has no attribute '%(attr)s'") % {
                            "class": self.__name__, "attr": name
                        }
                    )
                return name

    # Error strings
    class ErrorString:
        BAD_REQUEST = _('Bad request.')
        MISSING_STATUS = _('Status is missing.')
        INVALID_STATUS = _('Status is invalid.')
        MISSING_CLAIMED_ID = _('Claimed ID is missing.')
        MISSING_NONCE = _('Registration nonce is missing.')
        INVALID_NONCE = _('Registration nonce is invalid.')

# OpenID logging to django debug
import logging
logger = logging.getLogger('openid')
from openid import oidutil
oidutil.log = logger.warn
